@@ -11,15 +11,17 @@ class rsClient():
     def __init__(self, env, rs_iam_role):
         """
             When deploying, environment variables are injected depending on ENV via entrypoint.sh.
-            For local development, we fetch environemnt variables from local enviroment where we have variables for dev and prd, hence the except statement.
+            For local development, we fetch environemnt variables from local enviroment where we have variables for local, dev and prd, hence the except statement.
+
+            ! since we don't have a local redshift cluster we load into redshift dev 
         """
         try:
-            dbname=os.environ[f'REDSHIFT_DB']
-            user = os.environ[f'REDSHIFT_UNAME']
-            password = os.environ[f'REDSHIFT_PASSWORD']
-            host = os.environ[f'REDSHIFT_HOST']
-            port = os.environ[f'REDSHIFT_PORT']
-        except: 
+            dbname=os.environ['REDSHIFT_DB']
+            user = os.environ['REDSHIFT_UNAME']
+            password = os.environ['REDSHIFT_PASSWORD']
+            host = os.environ['REDSHIFT_HOST']
+            port = os.environ['REDSHIFT_PORT']
+        except Exception: 
             dbname=os.environ[f'REDSHIFT_DB_{env.upper()}']
             user = os.environ[f'REDSHIFT_UNAME_{env.upper()}']
             password = os.environ[f'REDSHIFT_PASSWORD_{env.upper()}']
@@ -70,7 +72,7 @@ class rsClient():
                 logging.error(f'Issue copying {file} into staging table. Exiting')
                 logging.error(f'message: {e}', exc_info=True)
                 sys.exit(1)
-        logging.info(f'Succesfully loaded all files in temp staging table')
+        logging.info('Succesfully loaded all files in temp staging table')
         return
 
     def _upsert(self, rs_target, rs_stage, upsert_keys):
@@ -102,11 +104,80 @@ class rsClient():
                 """)
             logging.info(f'UPSERT succesfull for {rs_stage}')
         except Exception as e:
-                logging.error(f'Issue UPSERTING stage into target. Exiting...')
+                logging.error('Issue UPSERTING stage into target. Exiting...')
                 logging.error(f'message: {e}', exc_info=True)
                 sys.exit(1)
     
+
+    def _upsert_non_lynx(self, rs_target, rs_stage, upsert_keys):
+        where_condition_target = ''
+        counter = 0
+        for col in upsert_keys:
+            if counter == 0:
+                line_target = f'{rs_target}.{col} = {rs_stage}.{col}'
+            else:
+                line_target = f'AND {rs_target}.{col} = {rs_stage}.{col}'
+            where_condition_target += str(line_target)
+            counter =+ 1
+        try:
+            self._query(f"""
+                    begin transaction;
+
+                    DELETE FROM {rs_target} 
+                    USING {rs_stage} 
+                    WHERE {where_condition_target} 
+                    ;
+                    
+                    INSERT INTO {rs_target}
+                    SELECT *
+                    FROM {rs_stage};
+
+                    DROP TABLE {rs_stage};
+                    end transaction;
+                """)
+            logging.info(f'UPSERT based on version succesfull for {rs_stage}')
+        except Exception as e:
+                logging.error('Issue UPSERTING stage into target. Exiting...')
+                logging.error(f'message: {e}', exc_info=True)
+                sys.exit(1)
+
+
+    def _upsert_version(self, rs_target, rs_stage, upsert_keys):
+        where_condition_target = ''
+        counter = 0
+        for col in upsert_keys:
+            if counter == 0:
+                line_target = f'{rs_target}.{col} = {rs_stage}.{col}'
+            else:
+                line_target = f'AND {rs_target}.{col} = {rs_stage}.{col}'
+            where_condition_target += str(line_target)
+            counter =+ 1
+        try:
+            self._query(f"""
+                    begin transaction;
+
+                    DELETE FROM {rs_target} 
+                    USING {rs_stage} 
+                    WHERE {where_condition_target} AND
+                        {rs_stage}.version > {rs_target}.version
+                    ;
+                    
+                    INSERT INTO {rs_target}
+                    SELECT *
+                    FROM {rs_stage};
+
+                    DROP TABLE {rs_stage};
+                    end transaction;
+                """)
+            logging.info(f'UPSERT based on version succesfull for {rs_stage}')
+        except Exception as e:
+                logging.error('Issue UPSERTING stage into target. Exiting...')
+                logging.error(f'message: {e}', exc_info=True)
+                sys.exit(1)
+    
+
     def _full_refresh(self, rs_target, rs_stage):
+        # delete is slow but it is not possible to do a truncate within an atomic transaction
         try:
             self._query(f"""
                     begin transaction;
@@ -123,7 +194,7 @@ class rsClient():
                 """)
             logging.info(f'FULL REFRESH succesfull for {rs_target}')
         except Exception as e:
-                logging.error(f'Issue REFRESHING target. Exiting...')
+                logging.error('Issue REFRESHING target. Exiting...')
                 logging.error(f'message: {e}', exc_info=True)
                 sys.exit(1)
     
@@ -131,13 +202,14 @@ class rsClient():
     def _append(self):
         None
 
-    def upsert(self, files, rs_target_schema, rs_target_table, upsert_keys):
+    def upsert(self, files, rs_target_schema, rs_target_table, upsert_keys, lynx=True, version=False):
         """
             ARGUMENTS
                 files: list of parquet files (eg. ['s3://bucket/folder/sub/file.parquet', ...])
                 rs_target_schema: redshift target schema (eg. 'ingest)
                 rs_target_table: redshift target table (eg. 'cogenius_xxx')
                 upsert_keys: list of fields to match records between the source (ie. stage) and target, in order to identify which records already exist
+                lynx: boolean if true, datecreated/datemodifed or version will be taken into account (depending on version argument). If false, only upsert keys are taken into account
                             
             RETURNS
                 None
@@ -152,7 +224,36 @@ class rsClient():
         rs_stage = f'{rs_target}_TEMP_STAGE'
         self._create_stage_like_target(rs_target, rs_stage)
         self._copy_parquet_into_stage(files, rs_stage)
-        self._upsert(rs_target, rs_stage, upsert_keys)
+        if lynx: 
+            self._upsert(rs_target, rs_stage, upsert_keys)
+        else:
+            self._upsert_non_lynx(rs_target, rs_stage, upsert_keys)
+        return
+
+
+    # REFACTOR - add to def upsert() with version boolean (lynx=true, version=true)
+    def upsert_version(self, files, rs_target_schema, rs_target_table, upsert_keys):
+        """
+            ARGUMENTS
+                files: list of parquet files (eg. ['s3://bucket/folder/sub/file.parquet', ...])
+                rs_target_schema: redshift target schema (eg. 'ingest)
+                rs_target_table: redshift target table (eg. 'cogenius_xxx')
+                upsert_keys: list of fields to match records between the source (ie. stage) and target, in order to identify which records already exist
+                            
+            RETURNS
+                None
+
+            ADDITIONAL INFO
+                List of upsert keys can be considered a composite key. If the composite key already exists within the target table, the associated record
+                will be deleted and replaced by the new record with the same composite key. 
+                Since and older partition can be reprocessed, an additional check is performed on version. Only if the version within the target is 
+                < than the version within the partition we're upserting, the record will be deleted from the target
+        """
+        rs_target = f'{rs_target_schema}.{rs_target_table}'
+        rs_stage = f'{rs_target}_TEMP_STAGE'
+        self._create_stage_like_target(rs_target, rs_stage)
+        self._copy_parquet_into_stage(files, rs_stage)
+        self._upsert_version(rs_target, rs_stage, upsert_keys)
         return
 
     def full_refresh(self,files, rs_target_schema, rs_target_table):
@@ -214,8 +315,18 @@ class rsClient():
         return
 
     def rs_to_df(self, dml):
-        df = pd.read_sql_query(dml, self.conn)
-        return df
+        return pd.read_sql_query(dml, self.conn)
+
+    
+    def rs_fetch_single_val(self, query):
+        result = None
+        with self.conn.cursor() as cur:
+            cur.execute(query)
+            result = cur.fetchone()
+            self.conn.commit()
+        return result[0]
+
+    
 
     def rs_column_to_api_query_param_list(self, rs_source_schema:str, rs_source_table:str, rs_source_column:str, api_query_params:str):
         """
@@ -300,4 +411,4 @@ class rsClient():
         s3.Object(s3_bucket, s3_key).put(Body=parquet_buffer.getvalue())
         logging.info(f'Stored {s3_key} on s3 {s3_bucket}')
         del df
-        return
\ No newline at end of file
+        return
